#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          25-06-2018
//      Created for Project :   FLAME
"""

from pyramid.httpexceptions import HTTPBadRequest, HTTPConflict, HTTPNotFound
from pyramid.view import view_defaults, view_config
from clmcservice.models import ServiceFunctionEndpoint
from clmcservice.whoamiapi.utilities import validate_sfendpoint_body


@view_defaults(renderer='json')
class WhoamiAPI(object):
    """
    A class-based view for accessing and mutating the configuration of SF endpoints - namely, the WHOAMI API.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='whoami_endpoints', request_method='GET')
    def get_all(self):
        """
        GET API call for all resources.

        :return: A list of all service function endpoint configurations found in the database.
        """

        return [instance.json for instance in ServiceFunctionEndpoint.query()]

    @view_config(route_name='whoami_endpoints_instance', request_method='GET')
    def get_one(self):
        """
        GET API call for a single resources.

        :return: One service function endpoint configuration instance retrieved from the database by querying the uniquely constrained columns.

        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_endpoint = self._get_sf_endpoint_from_url_string()
        if sf_endpoint is None:
            raise HTTPNotFound("A service function endpoint with the given URL query parameters doesn't exist.")
        else:
            return sf_endpoint.json

    @view_config(route_name='whoami_endpoints', request_method='POST')
    def post(self):
        """
        A POST API call to create a new service function endpoint.

        :return: A JSON response to the POST call - essentially with the data of the new resource

        :raises HTTPBadRequest: if request body is not a valid JSON for the configuration
        :raises HTTPConflict: if the unique constraints are not preserved after the creation of a new instance
        """

        # create an instance of the model and add it to the database table
        sf_endpoint = self._validate_and_create()
        json_data = sf_endpoint.json
        ServiceFunctionEndpoint.add(sf_endpoint)

        self.request.response.status = 201

        return json_data

    @view_config(route_name='whoami_endpoints_instance', request_method='PUT')
    def put(self):
        """
        A PUT API call to update a service function endpoint.

        :return: A JSON response representing the updated object

        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_endpoint = self._get_sf_endpoint_from_url_string()
        if sf_endpoint is None:
            raise HTTPNotFound("A service function endpoint with the given identifier doesn't exist.")
        else:
            try:
                body = self.request.body.decode(self.request.charset)
                validated_body = validate_sfendpoint_body(body)  # validate the content and receive a json dictionary object
            except AssertionError as e:
                raise HTTPBadRequest("Bad request content. Configuration format is incorrect: {0}".format(e.args))

            new_resource = validated_body
            old_resource = sf_endpoint.json
            updating = new_resource["sf_endpoint"] == old_resource["sf_endpoint"]

            if updating:
                ServiceFunctionEndpoint.delete(sf_endpoint)
                new_sf_endpoint = ServiceFunctionEndpoint(**validated_body)
                ServiceFunctionEndpoint.add(new_sf_endpoint)
            else:
                resource_exists = ServiceFunctionEndpoint.exists(new_resource["sf_endpoint"])
                if resource_exists:
                    raise HTTPConflict("Service function endpoint with this configuration already exists.")  # error 409 in case of resource conflict

                new_sf_endpoint = ServiceFunctionEndpoint(**validated_body)
                ServiceFunctionEndpoint.replace(sf_endpoint, new_sf_endpoint)

            return validated_body

    @view_config(route_name='whoami_endpoints_instance', request_method='DELETE')
    def delete(self):
        """
        Deletes an instance of a service function endpoint configuration in the database.

        :return: A content of the object that has been deleted

        :raises HTTPBadRequest: if the request parameters are invalid(invalid url query string)
        :raises HTTPNotFound: if a resource with the given parameters doesn't exist in the database
        """

        sf_endpoint = self._get_sf_endpoint_from_url_string()
        if sf_endpoint is None:
            raise HTTPNotFound("A service function endpoint with the given parameters doesn't exist.")
        else:
            deleted = sf_endpoint.json
            ServiceFunctionEndpoint.delete(sf_endpoint)
            return deleted

    def _get_sf_endpoint_from_url_string(self):
        """
        Retrieves a service function endpoint configuration from the database by validating and then using the request url parameters.

        :return: An instance of a service function endpoint configuration or None if not existing
        """

        sf_endpoint_id = self.request.params.get("sf_endpoint")
        if sf_endpoint_id is None:
            raise HTTPBadRequest("Request format is incorrect: sf_endpoint ID is not found in the URL query string.")

        sf_endpoint = ServiceFunctionEndpoint.get(sf_endpoint_id)

        return sf_endpoint

    def _validate_and_create(self):
        """
        Validates the request body and checks if a resource with the given attributes already exists.

        :return: a new instance of the model, if the resource doesn't exist
        :raises HTTPBadRequest: if request body is not a valid JSON for the configuration
        :raises HTTPConflict: if the unique constraints are not preserved after the creation of a new instance
        """

        try:
            body = self.request.body.decode(self.request.charset)
            validated_body = validate_sfendpoint_body(body)  # validate the content and receive a json dictionary object
        except AssertionError as e:
            raise HTTPBadRequest("Bad request content. Configuration format is incorrect: {0}".format(e.args))

        resource = validated_body

        resource_exists = ServiceFunctionEndpoint.exists(resource["sf_endpoint"])
        if resource_exists:
            raise HTTPConflict("Service function endpoint with the given identifier already exists.")  # error 409 in case of resource conflict

        # create an instance of the model
        sf_endpoint = ServiceFunctionEndpoint(**resource)

        return sf_endpoint
